// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include "qabstractopenxrgraphics_p.h"

QT_BEGIN_NAMESPACE

QAbstractOpenXRGraphics::QAbstractOpenXRGraphics()
{

}

void QAbstractOpenXRGraphics::setupWindow(QQuickWindow *)
{

}

QT_END_NAMESPACE
