// Copyright (C) 2021 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include "glsltype_p.h"

QT_BEGIN_NAMESPACE

using namespace GLSL;

Type::~Type()
{
}

QT_END_NAMESPACE
